import json
from .python_extractors import PytestExtractor, UnittestExtractor, PytestSCForgeExtractor
from .html_writer import write_html

DEFAULT = "default"


def extract(args):
    with open(args.file) as f:
        description = json.load(f)

    with open("question.html", "w") as f:
        write_html(description["rubric"], f)

    extractors = {
        "py": {
            "pytest": PytestExtractor,
            "unittest": UnittestExtractor,
            "pytest_scforge": PytestSCForgeExtractor,
            DEFAULT: UnittestExtractor
        }
    }

    if args.language in extractors:
        language_extractors = extractors[args.language]

        if args.framework in language_extractors:
            extractor = language_extractors[args.framework](description)
        elif DEFAULT in language_extractors:
            extractor = language_extractors[DEFAULT](description)
        else:
            raise Exception(
                f"No framework extractor found for {args.language} {args.framework}")
        extractor.write_source()
        extractor.write_tests()

    else:
        raise Exception(f"No extractor found for {args.language}")



import os

pytest_boilderplate = """
from solution import {0}
from showcode_forge import challenge, TestCase


@challenge(
    \"{2}, {3}\",
    [
        TestCase([], \"\", \"\"),
    ]
)
def test_{1}({2}, {3}):
    sut = {0}()
    assert sut.{1}({2}) == {3}
""".lstrip()

solution_bolierplate = """
class {0}:
    def {1}(self, {2}):
        pass


if __name__ == \"__main__\":
    print({0}().{1}( ))
""".lstrip()


def pytest_scaffold(args):
    class_name = args.class_name
    method_name = args.method_name
    argument_names = ', '.join(args.argument)
    result_name = args.result

    solution_path = os.path.join(args.output_dir, "solution.py")
    test_path = os.path.join(args.output_dir, "test_solution.py")

    with open(solution_path, "w") as f:
        f.write(solution_bolierplate.format(
            class_name,
            method_name,
            argument_names,
            result_name
        ))

    with open(test_path, "w") as f:
        f.write(pytest_boilderplate.format(
            class_name,
            method_name,
            argument_names,
            result_name
        ))



import os

question_boilerplate = """
<html>
    <head></head>
    <body>

    </body>
</html>
""".strip()


def scaffold_question(args):
    question_path = os.path.join(args.output_dir, "question.html")

    with open(question_path, "w") as f:
        f.write(question_boilerplate)



from .pytest_compiler import pytest_compiles
import json


def load_html(filename):
    with open(filename) as f:
        return f.read()


def compile(args):

    rubric = load_html(args.question_file)

    if args.language == "py" and args.framework == "pytest_scforge":
        data = pytest_compiles(args)
    else:
        raise Exception("Compilation is not yet supported")

    with open(args.output, "w") as f:
        json.dump({
            "title": "",
            "difficulty": 1,
            "className": "",
            "methodName": "",
            "rubric": rubric,
            **data,
        }, f, indent=4)



import os

from .pytest_scaffolding import pytest_scaffold
from .shared_scaffolding import scaffold_question


def scaffold(args):
    if not os.path.isdir(args.output_dir):
        os.makedirs(args.output_dir)

    if args.language == "py" and args.framework == "pytest_scforge":
        pytest_scaffold(args)

    scaffold_question(args)



def get_types(parameter):
    if isinstance(parameter, str):
        if(len(parameter) > 1):
            return [1]
        return [1, 2]
    if isinstance(parameter, bool):
        return [3]
    if isinstance(parameter, int):
        return [0, 5, 4]
    if isinstance(parameter, float):
        return [5, 4]

    if not isinstance(parameter, (list, tuple)):
        raise Exception("Unknown parameter type")

    if len(parameter) == 0:
        return [6, 7, 8, 9, 10, 11]
    if all(isinstance(p, int) for p in parameter):
        return [6, 11, 10]
    if all(isinstance(p, (float, int)) for p in parameter):
        return [11, 10]
    if all(isinstance(p, bool) for p in parameter):
        return [9]
    if all(isinstance(p, str) and len(p) <= 1 for p in parameter):
        return [8, 7]
    if all(isinstance(p, str) for p in parameter):
        return [7]

    raise Exception("Unknown parameter array type")



import os
import importlib
import importlib.util
import inspect
from functools import reduce
from contextlib import contextmanager
from ..shared import get_types

@contextmanager
def add_to_path(p):
    import sys
    old_path = sys.path
    sys.path = sys.path[:]
    sys.path.insert(0, p)
    try:
        yield
    finally:
        sys.path = old_path

def load_module(tested_file, load_as_main=False):
    with add_to_path(os.path.dirname(tested_file)):
        spec = importlib.util.spec_from_file_location(
            "__main__" if load_as_main else "tested_module", tested_file)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module


def get_test_cases(module):
    members = inspect.getmembers(module)
    test_case_members = [member[1]
                         for member in members if member[0] == "TestCase"]

    if len(test_case_members) != 1:
        raise Exception("Could not find test cases")

    return test_case_members[0].test_cases, test_case_members[0].parameters[0]

def build_parameter(name, inputs):
    possible_types = list(map(get_types, inputs))

    valid_types = list(reduce(lambda allowed, new: filter(lambda v: v in new, allowed), possible_types))
    if len(valid_types) < 1:
        raise Exception(f"Couldn't determine parameter type: {name}")
    return {
        "name": name,
        "type": valid_types[0]
    }
    

def build_parameters(test_cases, parameters):
    return [
        build_parameter(name.strip(), inputs)
        for name, inputs in zip(parameters, list(zip(*[t.inputs for t in test_cases])))]

def sum_points(test_cases):
    return sum(case.points for case in test_cases)

def get_return_type(test_cases):
    outputs = [t.output for t in test_cases]
    possible_types = list(map(get_types, outputs))
    valid_types = list(reduce(lambda allowed, new: filter(lambda v: v in new, allowed), possible_types))
    if len(valid_types) < 1:
        raise Exception("Couldn't determine output type")
    return valid_types[0]

def get_difficulty(points):
    if points < 10:
        return 0
    if points < 15:
        return 1
    return 2


def pytest_compiles(args):
    module = load_module(args.test_file)
    test_cases, parameters = get_test_cases(module)
    point_total = sum_points(test_cases)

    return {
        "points": point_total,
        "difficulty": get_difficulty(point_total),
        "returnType": get_return_type(test_cases),
        "parameters": build_parameters(test_cases, parameters.split(",")[:-1]),
        "unitTests": [case.to_dict() for case in test_cases]
    }



from .python_extractor import PythonExtractorABC


class PytestExtractor(PythonExtractorABC):

    def write_tests(self):
        with open(f"test_{self._method_name.lower()}.py", "w") as f:
            f.write("import pytest\n")
            f.write(f"from solution import {self._class_name}\n")
            f.write("\n\n")
            self._write_test(self._public_unit_tests, "public", f)
            f.write("\n\n")
            self._write_test(self._private_unit_tests, "hidden", f)

    def _write_test(self, tests, name, f):
        f.write(f'@pytest.mark.{name}\n')
        f.write(f"@pytest.mark.parametrize(\n")
        f.write(f"    \"{','.join(self._parameters)},result\",\n")
        f.write(f"    [\n")
        for test in tests:
            f.write(
                f"        ({', '.join(map(repr, [*test['inputs'], test['output']]))}),    # {test['description']}\n")
        f.write(f"    ]\n")
        f.write(f")\n")
        f.write(f"def test_{name}({', '.join(self._parameters)}, result):\n")
        f.write(f"    {self._class_instance_name} = {self._class_name}()\n")
        f.write(
            f"    assert {self._class_instance_name}.{self._method_name}({', '.join(self._parameters)}) == result\n")



import argparse
from .extractors import extract
from .compilers import compile
from .validator import validate
from .scaffolding import scaffold

parser = argparse.ArgumentParser()
subparsers = parser.add_subparsers()
parser.set_defaults(func=lambda *args: parser.print_help())

parser_extract = subparsers.add_parser("extract")
parser_extract.add_argument("file", type=str, help="Path to the challenge json file")
parser_extract.add_argument("--language", default="py", help="Which programming language should the code be generated in")
parser_extract.add_argument("--framework", default="unittest", help="Which test framework should the code be generated in")
parser_extract.set_defaults(func=extract)

parser_compile = subparsers.add_parser("compile")
parser_compile.add_argument("source_file", type=str, help="Path to the challenge source file")
parser_compile.add_argument("test_file", type=str, help="Path to the challenge test file")
parser_compile.add_argument("question_file", type=str, help="Path to the question file")
parser_compile.add_argument("--output", type=str, default="challenge.json", help="Path to the output json file")
parser_compile.add_argument("--language", default="py", help="Which programming language the source file is in")
parser_compile.add_argument("--framework", default="pytest_scforge", help="Which test framework the source test is in")
parser_compile.set_defaults(func=compile)

parser_validate = subparsers.add_parser("validate")
parser_validate.add_argument("file", type=str, help="Path to the challenge json file")
parser_validate.add_argument("--verbose", action="store_true", help="Enable additional logging")
parser_validate.set_defaults(func=validate)

parser_scaffold = subparsers.add_parser("scaffold")
parser_scaffold.add_argument("class_name", type=str, help="The name of the class in the solution file")
parser_scaffold.add_argument("method_name", type=str, help="The name of the invoked method")
parser_scaffold.add_argument("--argument", nargs="+", default=["arg1", "arg2"], help="The names of the method arguments")
parser_scaffold.add_argument("--result", type=str, default="result", help="The name of the result variable")
parser_scaffold.add_argument("--output_dir", type=str, default=".", help="The directory where the scaffold will be generated")
parser_scaffold.add_argument("--language", default="py", help="Which programming language the source file is in")
parser_scaffold.add_argument("--framework", default="pytest_scforge", help="Which test framework the source test is in")
parser_scaffold.set_defaults(func=scaffold)

def main():
    args = parser.parse_args()
    args.func(args)

if __name__ == "__main__":
    main()


import pathlib
from setuptools import setup, find_packages
import os

version = "0.7"

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()


dependencies = [
    "pytest"
]

setup(
    name='showcode_forge',
    packages=find_packages(),
    version=version,
    # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    license='mit',
    description='Tools for generating code from custom ShowCode Forge questions',
    author='Titusz Ban',
    author_email='tituszban@gmail.com',
    keywords=["showcode"],
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/tituszban/showcode_forge",
    install_requires=dependencies,
    setup_requires=dependencies,
    classifiers=[
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Development Status :: 4 - Beta',
        'Topic :: Software Development :: Build Tools',
        "License :: OSI Approved :: MIT License",
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
    ],
    entry_points={
        'console_scripts': [
            "showcode_forge=showcode_forge.__main__:main"
        ]
    }
)



from .python_extractor import PythonExtractorABC


class UnittestExtractor(PythonExtractorABC):

    def write_tests(self):
        with open(f"tests.py", "w") as f:
            f.write("import unittest\n")
            f.write(f"from solution import {self._class_name}\n")
            f.write("\n\n")
            self._write_test_class(self._public_unit_tests, "public", f)
            f.write("\n\n")
            self._write_test_class(self._private_unit_tests, "hidden", f)
            f.write("\n\n")
            f.write("if __name__ == '__main__':\n")
            f.write("    unittest.main()\n")

    def _write_test_class(self, tests, name, f):
        f.write(f"class {name.capitalize()}SolutionTest(unittest.TestCase):\n")
        for i, test in enumerate(tests):
            self._write_test_case(test, i, f)

    def _write_test_case(self, test, name, f):
        f.write(f"    def test_{name}(self):    # {test['description']}\n")
        f.write(
            f"        {self._class_instance_name} = {self._class_name}()\n")
        f.write(
            f"        self.assertEqual({self._class_instance_name}.{self._method_name}({', '.join(map(repr, test['inputs']))}), {test['output']})\n")
        f.write("\n")



import pytest


class TestCase:
    test_cases = []
    parameters = []

    def __init__(self, inputs, output, description, points=1, is_public=False):
        self.inputs = inputs
        self.output = output
        self.description = description
        self.points = points
        self.is_public = is_public

    def to_pytest_case(self):
        return pytest.param(
            *self.inputs, self.output,
            id=self.description or f"{self.inputs}:{self.output}"
        )

    def is_valid(self, fields):
        return len(fields.split(",")) == len(self.inputs) + 1

    def to_dict(self):
        return {
            "inputs": self.inputs,
            "output": self.output,
            "points": self.points,
            "isPublic": self.is_public,
            "description": self.description
        }


def challenge(fields, cases):
    assert all(isinstance(case, TestCase) for case in cases), \
        "Invalid case type"
    assert all(case.is_valid(fields) for case in cases), \
        "Invalid test case: wrong number of inputs"
    assert len(TestCase.parameters) == 0 or TestCase.parameters[0] == fields, \
        "Invalid test case: argument must always be the same"

    for case in cases:
        TestCase.test_cases.append(case)
    if len(TestCase.parameters) == 0:
        TestCase.parameters.append(fields)

    return pytest.mark.parametrize(fields, [case.to_pytest_case() for case in cases])



from abc import ABC, abstractmethod

class PythonExtractorABC(ABC):
    def __init__(self, description):
        self._class_name = description["className"]
        self._class_instance_name = self._class_name[0].lower()
        self._method_name = description["methodName"]
        self._parameters = [p["name"] for p in description["parameters"]]

        self._unit_tests = description["unitTests"]
        self._public_unit_tests = [t for t in self._unit_tests if t["isPublic"]]
        self._private_unit_tests = [t for t in self._unit_tests if not t["isPublic"]]

    def write_source(self):
        with open("solution.py", "w") as f:
            f.write(f"class {self._class_name}:\n")
            f.write(f"    def {self._method_name}(self, {', '.join(self._parameters)}):\n")
            f.write(f"        pass\n")
            f.write("\n\n")
            f.write("if __name__ == \"__main__\":\n")
            f.write(f"    {self._class_instance_name} = {self._class_name}()\n")
            f.write(f"    print({self._class_instance_name}.{self._method_name}())\n")
    
    @abstractmethod
    def write_tests(self):
        pass



import json
import os
from ..shared import get_types


class Validator:
    def __init__(self, args):
        self._args = args

    def _print_verbose(self, *args):
        if self._args.verbose:
            print(*args)

    def _check_json(self, path):
        self._print_verbose("Checking file")
        try:
            with open(path, encoding="utf-8") as f:
                return json.load(f)
        except json.decoder.JSONDecodeError as e:
            print(f"Invalid JSON: {e}")
            exit(1)

    def _check_required_fields(self, desc, required_fields):
        errors = []
        for name, t in required_fields:
            if name not in desc:
                errors.append(f"Missing field: '{name}'")
                continue

            if t is not None and type(desc[name]) != t:
                errors.append(
                    f"Field '{name}' has wrong type. Expected: '{t.__name__}', found: '{type(desc[name]).__name__}'")
                continue

            if (t == str or t == list) and len(desc[name]) <= 0:
                errors.append(f"Field '{name}' must not be empty")
                continue

        return errors

    def _check_fields(self, desc):
        self._print_verbose("Checking required fields")
        required_fields = [
            ("title", str),
            ("difficulty", int),
            ("points", int),
            ("rubric", str),
            ("className", str),
            ("methodName", str),
            ("returnType", int),
            ("parameters", list),
            ("unitTests", list)
        ]
        return self._check_required_fields(desc, required_fields)

    def _check_parameters(self, desc):
        self._print_verbose("Checking parameters")
        parameters = desc["parameters"]

        required_parameter_fields = [
            ("type", int),
            ("name", str)
        ]

        errors = []

        for i, parameter in enumerate(parameters):
            parameter_errors = self._check_required_fields(
                parameter, required_parameter_fields)
            for e in parameter_errors:
                errors.append(f"Parameter {i}: {e}")

            if parameter.get("type", 0) not in range(12):
                errors.append(
                    f"Parameter {i}: Type must be between 0 and 11. It is instead {parameter['type']}")

        return errors

    def _check_unit_tests(self, desc):
        self._print_verbose("Checking unit tests")
        unit_tests = desc["unitTests"]
        parameters = desc["parameters"]
        return_type = desc["returnType"]

        required_unit_test_fields = [
            ("inputs", list),
            ("output", None),
            ("isPublic", bool),
            ("description", str)
        ]

        errors = []

        for i, unit_test in enumerate(unit_tests):
            unit_test_errors = self._check_required_fields(
                unit_test, required_unit_test_fields)
            for e in unit_test_errors:
                errors.append(f"Unit test {i}: {e}")

            if "inputs" in unit_test:
                if len(parameters) != len(unit_test["inputs"]):
                    errors.append(
                        f"Unit test {i}: Wrong number of inputs. Number of parameters expected: {len(parameters)}, number of inputs: {len(unit_test['inputs'])}")
                    continue

                for j, inp in enumerate(unit_test["inputs"]):
                    param_type = parameters[j]["type"]
                    inp_types = get_types(inp)
                    if param_type not in inp_types:
                        errors.append(
                            "Unit test {0}, input {1}: Wrong type. Parameter requires type {2}, input is {3}".format(
                                i, j, param_type,
                                inp_types[0] if len(
                                    inp_types) <= 1 else f"one of {inp_types}"
                            ))

            if "output" in unit_test:
                output = unit_test["output"]
                output_types = get_types(output)

                if return_type not in output_types:
                    errors.append(
                        "Unit test {0}: Wrong output type. Return type is {1}, output is {2}".format(
                            i, return_type,
                            output_types[0] if len(
                                output_types) <= 1 else f"one of {output_types}"
                        )
                    )

        return errors

    def _check_difficulty(self, desc):
        self._print_verbose("Checking scoring and difficulty")
        unit_tests = desc["unitTests"]
        points = desc["points"]
        difficulty = desc["difficulty"]

        total_points = 0

        errors = []

        for i, unit_test in enumerate(unit_tests):
            if "points" not in unit_test:
                total_points += 1
                continue

            unit_test_points = unit_test["points"]

            if unit_test_points <= 0:
                errors.append(
                    f"Unit test {i}: Points must be greater than zero"
                )
                continue

            total_points += unit_test_points

        if points != total_points:
            errors.append(
                f"Invalid points: points are set to {points}, but unit tests have {total_points} points"
            )

        if difficulty not in range(3):
            errors.append(
                f"Invalid difficulty: {difficulty}. Difficulty must be between 0 and 2")
        else:
            if difficulty == 0 and points > 10:
                print(
                    f"Warning: difficulty is {difficulty} but points are {points}. The recommended points for Basic is up to 10")
            if difficulty == 1 and (points < 10 or points > 15):
                print(
                    f"Warning: difficulty is {difficulty} but points are {points}. The recommended points for Intermediate is between 10 and 15")
            if difficulty == 2 and points < 15:
                print(
                    f"Warning: difficulty is {difficulty} but points are {points}. The recommended points for Advanced is above 15")

        return errors

    def validate(self):
        self._print_verbose("Starting verification")

        path = self._args.file
        if not os.path.isfile(path):
            print(f"File not found: {path}")
            exit(1)

        desc = self._check_json(path)

        validators = (
            self._check_fields,
            self._check_parameters,
            self._check_unit_tests,
            self._check_difficulty
        )

        for validator in validators:
            errors = validator(desc)

            for error in errors:
                print(error)
            if any(errors):
                exit(1)

        self._print_verbose("Success")


def validate(args):
    validator = Validator(args)
    validator.validate()
